// Copyright (C) 2023 OWO Games - All Rights Reserved
#include "OWOLibrary.h"

FString UOWOLibrary::BuildSensation(FString name, int Frequency, float DurationInSeconds, int IntensityPercentage, float RampUp, float RampDown, float ExitDelay, FString muscle)
{
    uniquePtr<MicroSensation> micro = SensationsFactory::Create(Frequency, DurationInSeconds, IntensityPercentage, RampUp, RampDown, ExitDelay, !name.IsEmpty() ? ((std::string)TCHAR_TO_UTF8(*name)) : "");
    
    if (!muscle.IsEmpty()) {
        std::string sensationString = micro->ToString();
        FString withMuscle = BuildSensationWithMuscle(sensationString.c_str(), muscle);
        return withMuscle;
    }

    std::string sensationString = micro->ToString();
    return sensationString.c_str();
}

FString UOWOLibrary::ParseBakedSensation(owoString sensationString, owoString name, int id)
{
    uniquePtr<Sensation> sensationParsed = SensationsParser::Parse(sensationString);
    BakedSensation baked = BakedSensation(id, name, movePtr(sensationParsed), "0");
    return baked.Stringify().c_str();
}

FString UOWOLibrary::BakeSensation(FString sensation, FString name, int id) {
    std::string nameString = (std::string)TCHAR_TO_UTF8(*name);
    std::string sensationString = (std::string)TCHAR_TO_UTF8(*sensation);
#if WITH_EDITOR
    try {
        return ParseBakedSensation(sensationString, nameString, id);
    }
    catch (WrongSensationFormatException e) {
        std::string auxError = e.GetCode();
        FString error = auxError.c_str();
        UE_LOG(LogTemp, Error, TEXT("%s"),*error);
        GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("%s"), *error));
        return "ERROR";
    }
#else
    return ParseBakedSensation(sensationString, nameString, id);
#endif
}

FString UOWOLibrary::ChangeSensationIntensity(FString sensation, int intensity)
{
    if (!sensation.Contains("|")) {
        UE_LOG(LogTemp, Error, TEXT("The sensation doesnt have any muscles attached!"));
        return "100,1,100,0,0,0,";
    }

    std::string sensationString = (std::string)TCHAR_TO_UTF8(*sensation);
    uniquePtr<SensationWithMuscles> sensationParsed = SensationWithMusclesParser::Parse(sensationString);
    
    sensationParsed->muscles = sensationParsed->muscles.WithIntensity(intensity);

    return  sensationParsed->ToString().c_str();
}

sharedPtr<Sensation> UOWOLibrary::ParseSensation(owoString sensationString)
{
    return SensationsParser::Parse(sensationString);
}

FString UOWOLibrary::BuildSensationChain(FString A, FString B)
{
    TArray<FString> sensations = { A, B, };
    owoVector<sharedPtr<Sensation>> sensationsV;
    
    for (int i = 0; i < sensations.Num(); i++) {
        sharedPtr<Sensation> sensation = nullptr;
        std::string sensationString = (std::string)TCHAR_TO_UTF8(*sensations[i]);
#if WITH_EDITOR
        try {
            sensation = ParseSensation(sensationString);
        } 
        catch (WrongSensationFormatException e) {
            std::string auxError = e.GetCode();
            FString error = auxError.c_str();
            UE_LOG(LogTemp, Error, TEXT("%s"),*error);
            GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("%s"), *error));
        }
#else
        sensation = ParseSensation(sensationString);
#endif
        sensationsV.push_back(sensation);

    }

    SensationsSequence* sequence = new SensationsSequence(sensationsV);

    std::string sequenceString = sequence->ToString();
    delete sequence;
    return sequenceString.c_str();
}

FString UOWOLibrary::BuildMuscle(EOWOMusclesEnum muscle, int IntensityPercentage)
{
    return (new Muscle((int)muscle, IntensityPercentage))->ToString().c_str();
}

FString UOWOLibrary::BuildMuscleGroupFromArray(TArray<FString> muscles) {
    FString muscleString = muscles[0];
    for (int i = 1; i < muscles.Num(); i++) {
        muscleString += "," + muscles[i];
    }

    MusclesGroup group = MusclesParser::Parse((std::string)TCHAR_TO_UTF8(*muscleString));
    return group.ToString().c_str();
}

FString UOWOLibrary::BuildMuscleGroup(FString A, FString B)
{
    TArray<FString> muscles = { A, B, };

    return UOWOLibrary::BuildMuscleGroupFromArray(muscles);
}

FString UOWOLibrary::ImportSensation(UImportedSensation* sensation) {
    if (sensation == nullptr) return "";
    return sensation->SensationsData;
}

FString UOWOLibrary::ImportGameAuth(FString gameId, UImportedAuth* auth) {
    if (auth == nullptr) return "";
    return gameId + "*AUTH*" + auth->SensationsData;
}

owoString UOWOLibrary::AddSensationGroup(owoString muscles, sharedPtr<Sensation> sensation)
{
    MusclesGroup musclesGroup = MusclesParser::Parse(muscles);
    SensationWithMuscles sensationM = SensationWithMuscles(sensation, musclesGroup);
    return sensationM.ToString();
}

Muscle UOWOLibrary::GetMuscle(std::vector<std::string > parameters)
{
    return  Muscle(stoi(parameters[0]), stoi(parameters[1]));;
}

FString UOWOLibrary::BuildSensationWithMuscle(FString sensationS, FString muscle)
{
    std::string sensationString = std::string(TCHAR_TO_UTF8(*sensationS));
    sharedPtr<Sensation> sensation;
#if WITH_EDITOR
    try {
        sensation =  ParseSensation(sensationString);
    }
    catch (WrongSensationFormatException e) {
        std::string auxError = e.GetCode();
        FString error = auxError.c_str();
        UE_LOG(LogTemp, Error, TEXT("%s"),*error);
        GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("%s"), *error));
        return "ERROR";
    }
#else
        sensation = ParseSensation(sensationString);
#endif
    std::string muscleString = std::string(TCHAR_TO_UTF8(*muscle));
    std::string result;
    if (muscle.IsEmpty())
    {
        std::vector<Muscle> muscles;
        muscles.push_back(Muscle::Pectoral_R());
        sensation = sensation->WithMuscles(muscles);
    }
    else {
        if (muscle.Contains(",")) {
#if WITH_EDITOR
            try {
                result = AddSensationGroup(muscleString, sensation);
            }
            catch (WrongMusclesFormatException e) {
                std::string auxError = e.GetCode();
                FString error = auxError.c_str();
                UE_LOG(LogTemp, Error, TEXT("%s"),*error);
                GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("%s"), *error));
            }
#else
            result = AddSensationGroup(muscleString, sensation);
#endif
            return result.c_str();
        }
        else {
            std::vector<Muscle> muscles;
            std::vector<std::string> parameters = OWOGame::String::Split(muscleString, '%');

#if WITH_EDITOR
                try {
                    muscles.push_back(GetMuscle(parameters)); 
                }
                catch (...) {
                    WrongMusclesFormatException e;
                    std::string auxError = e.GetCode();
                    FString error = auxError.c_str();
                    UE_LOG(LogTemp, Error, TEXT("%s"), *error);
                    GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("%s"), *error));
                }
#else
            muscles.push_back(GetMuscle(parameters));
#endif
            
            sensation = sensation->WithMuscles(muscles);
        }      
    }
    result = sensation->ToString();
    return result.c_str();
}

FString UOWOLibrary::Ball()
{
    return BuildSensation("", 100, 0.2f, 90);
}
FString UOWOLibrary::Dart()
{
    return BuildSensation("", 10, 0.1f, 30);
}

FString UOWOLibrary::Punch()
{
    return BuildSensation("", 100, 0.8f, 100, 0, 0.8f);
}

FString UOWOLibrary::Grip()
{
    return BuildSensation("", 60, 0.1f, 70);
}

FString UOWOLibrary::FreeFall()
{
    return BuildSensation("", 100, 2.5f, 60);
}

FString UOWOLibrary::InsectBites()
{
    return BuildSensation("", 13, 1.0f, 80);
}

FString UOWOLibrary::ShootEntry()
{
    return BuildSensation("", 30, 0.1f, 100);
}

FString UOWOLibrary::ShootExit()
{
    return BuildSensation("", 20, 0.1f, 100);
}

FString UOWOLibrary::Bleeding()
{
    return BuildSensation("", 100, 0.1f, 60);
}

FString UOWOLibrary::MachineGunRecoil()
{
    return BuildSensation("", 11, 1.0f, 100);
}

FString UOWOLibrary::OneHandRecoil()
{
    return BuildSensation("", 100, 0.1f, 100);
}

FString UOWOLibrary::Dagger()
{
    return BuildSensationChain(BuildSensation("", 60, 0.1f, 100), BuildSensation("", 100, 2.0f, 100, 0.3f, 2.0f));
}

FString UOWOLibrary::Shoot()
{
    return BuildSensationChain(BuildSensation("", 30, 0.1f, 100), BuildSensation("", 100, 0.1f, 60));
}

FString UOWOLibrary::Heartbeat()
{
    return BuildSensationWithMuscle(BuildSensationChain(BuildSensation("", 100, 0.1f, 70, 0, 0, 0.1f), BuildSensation("", 100, 0.2f, 90, 0, 0.1f)), BuildMuscle(EOWOMusclesEnum::Pectoral_L, 100));
}

FString UOWOLibrary::Hug()
{
    return BuildSensationChain(BuildSensationChain(BuildSensation("", 100, 3.0f, 60, 1.5f, 0.5f), BuildSensation("", 100, 0.2f, 70)), BuildSensation("", 100, 0.3f, 40));
}

FString UOWOLibrary::AllMuscles()
{
    return BuildMuscleGroup(BuildMuscleGroup(BuildMuscleGroup(BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Pectoral_R, 100), BuildMuscle(EOWOMusclesEnum::Pectoral_L, 100)), BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Abdominal_L, 100), BuildMuscle(EOWOMusclesEnum::Abdominal_R, 100))), BuildMuscleGroup(BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Arm_L, 100), BuildMuscle(EOWOMusclesEnum::Arm_R, 100)), BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Dorsal_L, 100), BuildMuscle(EOWOMusclesEnum::Dorsal_R, 100)))), BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Lumbar_L, 100), BuildMuscle(EOWOMusclesEnum::Lumbar_R, 100)));
}

FString UOWOLibrary::FrontMuscles()
{
    return BuildMuscleGroup(BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Pectoral_R, 100), BuildMuscle(EOWOMusclesEnum::Pectoral_L, 100)), BuildMuscleGroup(BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Abdominal_R, 100), BuildMuscle(EOWOMusclesEnum::Abdominal_L, 100)), BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Arm_R, 100), BuildMuscle(EOWOMusclesEnum::Arm_L, 100))));
}

FString UOWOLibrary::FrontMusclesWithoutArms()
{
    return BuildMuscleGroup(BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Pectoral_R, 100), BuildMuscle(EOWOMusclesEnum::Pectoral_L, 100)), BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Abdominal_R, 100), BuildMuscle(EOWOMusclesEnum::Abdominal_L, 100)));
}

FString UOWOLibrary::Pectorals()
{
    return BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Pectoral_R, 100), BuildMuscle(EOWOMusclesEnum::Pectoral_L, 100));
}

FString UOWOLibrary::Abdominals()
{
    return BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Abdominal_R, 100), BuildMuscle(EOWOMusclesEnum::Abdominal_L, 100));
}

FString UOWOLibrary::Arms()
{
    return BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Arm_R, 100), BuildMuscle(EOWOMusclesEnum::Arm_L, 100));
}

FString UOWOLibrary::BackMuscles()
{
    return BuildMuscleGroup(BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Dorsal_R, 100), BuildMuscle(EOWOMusclesEnum::Dorsal_L, 100)), BuildMuscleGroup(BuildMuscle(EOWOMusclesEnum::Lumbar_R, 100), BuildMuscle(EOWOMusclesEnum::Lumbar_L, 100)));
}
